
/**********************************************************************
 * $Id: command.c,v 1.4 93/01/15 13:04:19 drew Exp $
 **********************************************************************/

/**********************************************************************
 *   Copyright 1990,1991,1992,1993 by The University of Toronto,
 *		      Toronto, Ontario, Canada.
 * 
 *			 All Rights Reserved
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its  documentation for  any purpose is  hereby granted  without
 * fee, provided that the above copyright notice appears in all copies
 * and  that both the  copyright notice  and   this  permission notice
 * appear in   supporting documentation,  and  that the  name  of  The
 * University  of Toronto  not  be  used in  advertising or  publicity
 * pertaining   to  distribution   of  the  software without specific,
 * written prior  permission.   The  University of   Toronto makes  no
 * representations  about  the  suitability of  this software  for any
 * purpose.  It  is    provided   "as is"  without express or  implied
 * warranty.
 *
 * THE UNIVERSITY OF TORONTO DISCLAIMS  ALL WARRANTIES WITH REGARD  TO
 * THIS SOFTWARE,  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS, IN NO EVENT SHALL THE UNIVERSITY OF TORONTO  BE LIABLE
 * FOR ANY  SPECIAL, INDIRECT OR CONSEQUENTIAL  DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR  PROFITS, WHETHER IN
 * AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING
 * OUT  OF  OR  IN  CONNECTION WITH  THE  USE  OR PERFORMANCE  OF THIS
 * SOFTWARE.
 *
 **********************************************************************/

#include <xerion/useful.h>
#include <xerion/simulator.h>
#include <xerion/commands.h>

#include "unit.h"
#include "ebp.h"


#define expertIsGated(e)	(Mgate(e))

static int	resetGate	ARGS((Group gate, Group output)) ;
static int	gateExpert	ARGS((Group gate, Group expert)) ;
static int	ungateExpert	ARGS((Group expert)) ;
static Unit	findGateUnit	ARGS((Group gateGroup, Group expertGroup)) ;

/***********************************************************************
 *	Name:		command_gate
 *	Description:	gate a set of expert groups
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_gate - 0 on failure, 1 on success
 ***********************************************************************/
int	command_gate(tokc, tokv)
  int	tokc ;
  char	**tokv ;
{
  Group	gate, output, expert ;
  char	*name ;

  IUsage("<gate> <ouput> <expert1> [<expert2> ...]") ;
  if (GiveHelp(tokc)) {
    ISynopsis("gate a set of experts");
    IHelp
      (IHelpArgs,
       "This  command is used to build a  mixture of experts network. To use",
       "it, you must have already created the following groups:",
       "",
       "n expert groups with m units:",
       "	bp-> addGroup \"Expert 1\" <m>",
       "	...",
       "	bp-> addGroup \"Expert <n>\" <m>",
       "",
       "a dummy output group with m units, no input connections and type",
       "GATED:",
       "	bp-> addGroup -type OUTPUT -type GATED \"Target\" <m>",
       "	bp-> disconnectGroups '.*' 'Target'",
       "",
       "a GATE unit with n units (the same number of units as experts)",
       "	bp-> addGroup -type GATE \"Gate\" <n>",
       "",
       "With these groups you can then use the command:",
       "	bp-> gate \"Gate\" \"Target\" \"Expert 1\" ... \"Expert <n>\"",
       "",
       "The command  will  associate each unit  in the gate with one of  the",
       "expert groups (in the order specified on the command line), and  all",
       "of the  experts with the output  group. When an example is presented",
       "to  the network, the targets are stored  in the output  group.   The",
       "error measure is then calculated using the gate and experts, not the",
       "output group. ",
       "",
       "The major  assumption is that the errors  of each expert come from a",
       "gaussian distribution with  0 mean, and fixed variance.  The gate is",
       "used to calculate mixing  proportions for each of these gaussians to",
       "produce  a mixture  of gaussians error function  for the entire net.",
       "The error  being  optimized is the -log  of the probability that the",
       "observed  errors  came  from  the  underlying mixture  of  gaussians",
       "distribution.",
       "",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  /* Parse any command line options */
  name = *tokv ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }
  if (tokc < 3)
    IErrorAbort(IPrintUsage(name, usage)) ;

  gate = groupFromName(currentNet, *tokv) ;
  if (gate == NULL)
    IErrorAbort("Unknown group: \"%s\".", *tokv) ;
  if (!(gate->type & GATE))
    IErrorAbort("\"%s\" is not a gate group.", *tokv) ;

  ++tokv, --tokc ;
  output = groupFromName(currentNet, *tokv) ;
  if (output == NULL)
    IErrorAbort("Unknown group: \"%s\".", *tokv) ;
  if (!(output->type & GATED))
    IErrorAbort("\"%s\" is not a gated group.", *tokv) ;

  if (gate->numUnits != tokc - 1) {
    IError("Different number of gating units and experts (%d and %d)",
	   gate->numUnits, tokc - 1) ;
    return 0 ;
  }

  resetGate(gate, output) ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    expert = groupFromName(currentNet, *tokv) ;
    if (expert == NULL)
      IErrorAbort("Unknown group: \"%s\".", *tokv) ;

    if ((output->type & ENSEMBLEERROR) && !(expert->type & ENSEMBLETRANSFER)) {
      IError("Output group \"%s\" is ensemble, but expert \"%s\" isn't",
	     output->name, expert->name) ;
      return 0 ;
    }
    if (!(output->type & ENSEMBLEERROR) && (expert->type & ENSEMBLETRANSFER)) {
      IError("Output group \"%s\" isn't ensemble, but expert \"%s\" is",
	     output->name, expert->name) ;
      return 0 ;
    }

    if (expertIsGated(expert))
      ungateExpert(expert) ;
    gateExpert(gate, expert) ;
  }

  return 1 ;
}  
/********************************************************************/
    

/*********************************************************************
 *	Name:		resetGate
 *	Description:	
 *	Parameters:
 *	  Group	gate - 
 *	Return Value:
 *	  int	resetGate - 
 *********************************************************************/
static int	resetGate(gate, output)
  Group		gate ;
  Group		output ;
{
  int	idx ;

  if (!(gate->type & GATE)
      || !(output->type & OUTPUT) || !(output->type & GATED))
    return 0 ;

  Mgated(gate)  = output ;
  Mgate(output) = gate ;

  if (Mmixture(gate))
    MMdestroy(Mmixture(gate)) ;

  Mmixture(gate) = createMixture(gate->numUnits, 0.0, 0.0) ;
  for (idx = 0 ; idx < gate->numUnits ; ++idx) {
    Gaussian	gaussian = MMgaussian(Mmixture(gate), idx) ;
    MGsetVariance(gaussian, 1.0) ;
    Mgaussian(gate->unit[idx]) = gaussian ;
  }

  for (idx = 0 ; idx < gate->numUnits ; ++idx) {
    Unit	unit = gate->unit[idx] ;
    if (Mexpert(unit))
      ungateExpert(Mexpert(unit)) ;
    Mexpert(unit) = NULL ;
  }

  return 1 ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		gateExpert
 *	Description:	
 *	Parameters:
 *	  Group	gate - 
 *	  Group	expert - 
 *	Return Value:
 *	  int	gateExpert - 
 *********************************************************************/
static int 
gateExpert(gate, expert)
  Group	gate ;
  Group	expert ;
{
  Unit		gateUnit ;

  if (!(gate->type & GATE))
    return 0 ;

  if (expert->numUnits != Mgated(gate)->numUnits)
    return 0 ;

  gateUnit = findGateUnit(gate, NULL) ;
  if (gateUnit == NULL)
    return 0 ;

  Mexpert(gateUnit) = expert ;
  Mgate(expert)     = gate ;

  return 1 ;
}
/********************************************************************/


/*********************************************************************
 *	Name:		ungateExpert
 *	Description:	
 *	Parameters:
 *	  Group	expert - 
 *	Return Value:
 *	  int	ungateExpert - 
 *********************************************************************/
static int
ungateExpert(expert)
  Group	expert ;
{
  Group		gate = Mgate(expert) ;
  Unit		gateUnit ;

  if (!(gate->type & GATE))
    return 0 ;

  gateUnit = findGateUnit(gate, expert) ;

  if (gateUnit == NULL)
    return 0 ;

  Mexpert(gateUnit) = NULL ;
  Mgate(expert)     = NULL ;

  return 1 ;
}
/********************************************************************/


/********************************************************************/
static Unit
findGateUnit(gateGroup, expertGroup)
  Group	gateGroup ;
  Group	expertGroup ;
{
  int	idx ;

  for (idx = 0 ; idx < gateGroup->numUnits ; ++idx)
    if (Mexpert(gateGroup->unit[idx]) == expertGroup)
      return gateGroup->unit[idx] ;

  return NULL ;
}
/********************************************************************/


/* ARGSUSED */
int command_forceRedraw(tokc, tokv)
int tokc;
char *tokv[];
{
  int *p;
  if (GiveHelp(tokc)) {
    IUsage(" ");
    ISynopsis("Forces all displays to redraw");
    IHelp(IHelpArgs,
	  NULL);
    return 1;
  }

  catchupOnDisplayResets() ;
  if (din != stdin)
    delayDisplayResets() ;
  return 1 ;
}

String		linkLongName ARGS((Net, String	name)) ;
/***********************************************************************
 *	Name:		command_linkLongName
 *	Description:	returns the long name of a unit (given short
 *			name (e.g. "Bias" -> "currentNet.group[0].unit[0].")
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_createUnit - 0 on failure, 1 on success
 ***********************************************************************/
int
command_linkLongName (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char	*name, *linkName, *longName ;

  IUsage("<name>") ;
  if (GiveHelp(tokc)) {
    ISynopsis("return the long name of a link.") ;
    IHelp
      (IHelpArgs,
       "Returns the  long  name of a ling, given the link's real name. For",
       "example,",
       "",
       "\txerion-> linkLongName \"Input.0 -> Hidden.1\"",
       "",
       "would return the string:",
       "\tcurrentNet.links[34]",
       "SEE ALSO",
       "unitLongName, groupLongName",
       NULL);
    return 0;
  }
  
  name = tokv[0] ;
  if (tokc != 2)
    IErrorAbort(IPrintUsage(name, usage)) ;
  linkName = tokv[1] ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.") ;

  longName = linkLongName(currentNet, linkName) ;

  if (longName) {
    fprintf(dout, "%s\n", longName) ;
    free((void *)longName) ;
  } else {
    IErrorAbort("Unknown link: \"%s\"", linkName) ;
  }
  return 1 ;
}
/**********************************************************************/
String 
linkLongName(net, name)
  Net		net ;
  String	name ;
{
  char	longName[128] ;
  int	idx ;

  if (net == NULL || name == NULL)
    return NULL ;

  for (idx = 0 ; idx < net->numLinks ; ++idx) {
    if (strcmp(name, net->links[idx]->name) == 0)
      break ;
  }

  if (idx == net->numLinks)
    return NULL ;
  
  sprintf(longName, "currentNet.links[%d]", idx) ;

  return strdup(longName) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		command_relativeError
 *	Description:	returns the long name of a unit (given short
 *			name (e.g. "Bias" -> "currentNet.group[0].unit[0].")
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_createUnit - 0 on failure, 1 on success
 ***********************************************************************/
typedef struct {
  int	n ;
  Real	sumTarget ;
  Real	sumTargetSq ;
  Real	sumErrorSq ;
} ErrorStats ;
#define SQR(x)	((x)*(x))
static void	sumStats ARGS((Unit unit, void *data)) ;
/**********************************************************************/
static void 
sumStats(unit, data)
  Unit		unit ;
  void		*data ;
{
  ErrorStats	*stats = (ErrorStats *)data ;
  Real		target, output ;
  
  stats->n++ ;
  stats->sumTarget   += unit->target ;
  stats->sumTargetSq += SQR(unit->target) ;
  stats->sumErrorSq  += SQR(unit->target - unit->output) ;
}
/**********************************************************************/
int 
command_relativeError (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  char		*name ;
  int		setMask ;
  ExampleSet	exampleSet ;
  int		oldIdx, idx ;
  ErrorStats	stats ;
  String	format = "%g" ;

  IUsage("[-format <format>] [ -type <type>]");
  if (GiveHelp(tokc)) {
    ISynopsis("compute the relative error on an example set.") ;
    IHelp
      (IHelpArgs,
       NULL);
    return 0;
  }
  
  /***********************
   * Parse the command line options 
   */
  name = *tokv ;
  setMask = 0 ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-type", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      setMask = findClassMask(EXAMPLESET_CLASS, *tokv) ;
      if (setMask == 0)
	IErrorAbort("Unknown example set: \"%s\".", *tokv) ;
    } else if (strncmp(*tokv, "-format", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      format = *tokv ;
    } else if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }
  if (setMask == 0)
    setMask = TRAINING ;

  if (currentNet == NULL)
    IErrorAbort("There is no current net.");
   
  exampleSet = netGetExampleSet(currentNet, setMask) ;
  if (exampleSet == NULL)
    IErrorAbort("Unknown example set type: \"0x%0x\".", setMask) ;

  if (exampleSet->numExamples == 0)
    IErrorAbort("Example Set is empty") ;

  if (tokc != 0)
    IErrorAbort(IPrintUsage(name, usage)) ;

  /***********************
   * Now, do the examples 
   */
  stats.sumTarget	= 0.0 ;
  stats.sumTargetSq	= 0.0 ;
  stats.sumErrorSq	= 0.0 ;
  stats.n		= 0 ;

  oldIdx = exampleSet->currentIdx ;
  for (idx = 0 ; idx < exampleSet->numExamples ; ++idx) {
    MgetNumber(exampleSet, idx) ;
    MupdateNetActivities(currentNet) ;
    netForAllUnits(currentNet, OUTPUT, sumStats, &stats) ;
  }
  exampleSet->currentIdx = oldIdx ;

  fprintf(dout, format, stats.sumErrorSq/(stats.sumTargetSq 
					  - SQR(stats.sumTarget)/stats.n)) ;
  fputc('\n', dout) ;

  return 1 ;
}
/**********************************************************************/
