#include <math.h>
#include <xerion/useful.h>
#include <xerion/simulator.h>
#include <xerion/sigmoid.h>
#include <xerion/commands.h>

#include "unit.h"
#include "ebp.h"
#include "ensemble.h"
#include "ensembleCost.h"
#include "sigmoid.h"


#define ONE_OVER_SQRT_2PI 0.398942280401432677939946059935
/**********************************************************************/
#define expertIsGated(e)	(Mgate(e))
/**********************************************************************/

static void	ensembleDotProductCombIn 	ARGS((Unit	unit)) ;
static void	ensembleDotProductCombBack	ARGS((Unit	unit)) ;

static void   ensembleTransfer	   ARGS((Unit	unit)) ;
static void   ensembleTransferBack ARGS((Unit	unit)) ;

static void	ensembleSumSquareError	       ARGS((Unit	unit)) ;
static void	ensembleSumSquareErrorDeriv    ARGS((Unit	unit)) ;

static Real	square		ARGS((double  x)) ;
static Real	adjustedTarget	ARGS((const Unit	unit)) ;

#ifndef M_SQRT2
#define M_SQRT2 1.414
#endif  /* dont forget to fix this later */

/**********************************************************************/
static Real	
square(x)
  double x ;
{
  return (Real) (x * x) ;
}
/**********************************************************************/

/**********************************************************************/
static Real	adjustedTarget(unit) 
  const	Unit	unit ;
{
  Real	target = unit->target ;
  Real	radius = MzeroErrorRadius(unit->net) ;

  if (radius > 0.0) {
    Real	upper = target + radius ;
    Real	lower = target - radius ;
    if (unit->output > upper)
      target = upper ;
    else if (unit->output < lower)
      target = lower ;
    else
      target = unit->output ;
  }

  return target ;
}
/**********************************************************************/

/**********************************************************************/
void
setEnsembleCombIn(group)
  Group	group ;
{
  McombInProc(group)   = ensembleDotProductCombIn ;
  McombBackProc(group) = ensembleDotProductCombBack ;
}
/**********************************************************************/

/**********************************************************************/
void 
setEnsembleTransfer(group)
  Group	group ;
{
  setStandardForwardBackward(group) ;
  MtransferProc(group)	   = ensembleTransfer ;
  MtransferBackProc(group) = ensembleTransferBack ;
}
/**********************************************************************/

/**********************************************************************/
void 
setEnsembleError(group)
  Group	group ;
{
  McontributeErrorProc(group) 	  = ensembleSumSquareError ;
  McontributeErrorBackProc(group) = ensembleSumSquareErrorDeriv ;

  MerrorStdDevVar(group) = createVariable("ErrorStdDeviation", 
					  group->net, LOG_TRANSFORM) ;
  variableSetValue(MerrorStdDevVar(group), 1.0) ;
}
/**********************************************************************/


/***********************************************************************
 *	Name:		ensembleDotProductCombIn
 *	Description:	computes the mean and variance of the total input
 *                      to a unit. Different code for HIDDEN and OUTPUT.
 *	Parameters:	
 *		Unit	unit - the unit to calculate the mean and variance for.
 *	Return Value:	
 *                      NONE
 ***********************************************************************/
static void 
ensembleDotProductCombIn(unit) 
  Unit		unit ;
{
  Link		*incoming   = unit->incomingLink ;
  unsigned int	numIncoming = unit->numIncoming ;
  Real		totalInput, totalVar;
  unsigned int	idx ;

  totalInput = 0.0 ;
  totalVar = 0.0 ;

  if (unit->group->type & OUTPUT || expertIsGated(unit->group)) {
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link        link = incoming[idx] ;
      Real        preMean = link->preUnit->output;
      Real        preVar  = MoutputVar(link->preUnit);
      Real        linkMean = MgetWeight(link) ;
      Real        linkVar  = MgetWeightVar(link) ;

      totalInput += linkMean * preMean;
      totalVar
	+= linkMean * linkMean * preVar 
	  + preMean * preMean * linkVar
	    + linkVar * preVar;
    }
  } else {
    /* if its not an output unit it must be in the single hidden layer */
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link        link = incoming[idx] ;
      Real        preMean = link->preUnit->output;
      totalInput += MgetWeight(link) * preMean;
      totalVar   += MgetWeightVar(link) * preMean * preMean;
    }
  }
  unit->totalInput = totalInput ;
  MinputVar(unit)  = totalVar ;
}
/***********************************************************************
 *	Name:		ensembleDotProductCombBack
 *	Description:	back propagates derivates through the incoming
 *			links of a unit and adds them to the pre units'
 *			outputDerivs and the links derivs.
 *	Parameters:	
 *		Unit	unit - the unit to back propagate from
 *	Return Value:	NONE
 ***********************************************************************/
static void
ensembleDotProductCombBack(unit) 
  Unit		unit ;
{
  Link		*incoming   = unit->incomingLink ;
  Real		inputDeriv  = unit->inputDeriv ;
  Real		inputVarDeriv  = MinputVarDeriv(unit);
  unsigned int	numIncoming = unit->numIncoming ;
  unsigned int	idx ;

  if (unit->group->type & OUTPUT || expertIsGated(unit->group)) {
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link	link    = incoming[idx] ;
      Unit	preUnit = link->preUnit ;
      Real	preMean = preUnit->output;
      Real	preVar  = MoutputVar(preUnit);
      Real	linkMean = MgetWeight(link) ;
      Real	linkVar  = MgetWeightVar(link) ;
  
      preUnit->outputDeriv 
	+= inputDeriv * linkMean + 2 * inputVarDeriv * preMean * linkVar;
      MoutputVarDeriv(preUnit) 
	+= inputVarDeriv * (linkMean * linkMean + linkVar);

      MaddToWeightDeriv(link,
			inputDeriv*preMean + 2*inputVarDeriv*linkMean*preVar);
      MaddToWeightVarDeriv(link, 
			   inputVarDeriv * (preMean * preMean + preVar));
    }
  } else { 
    /* we can assume that the variance of the preUnit is 0 */
    for (idx = 0 ; idx < numIncoming ; ++idx) {
      Link	link    = incoming[idx] ;
      Unit	preUnit = link->preUnit ;
      Real	preMean = preUnit->output;
  
      MaddToWeightDeriv   (link, inputDeriv * preMean) ;
      MaddToWeightVarDeriv(link, inputVarDeriv * preMean * preMean);
    }
  }
}
/**********************************************************************/

/*********************************************************************
 *	Name:		ensembleTransfer(Unit)
 *	Description:	Passes a hidden unit's totalInput through a logistic
 *                      and an output unit's through a linear transform 
 *			and returns the new value.
 *			ensembleTransferBack returns the derivative of
 *			the transformation.
 *                      Both functions SET the relevant variance fields.
 *	Parameters:
 *	         Unit	unit - the unit (unchanged)
 *	Return Value:
 *	  static void	logisticTransfer - the transformed value
 *********************************************************************/
static void 
ensembleTransfer(unit)
    Unit	unit ;
{
    Real        output;
    if (unit->group->type & OUTPUT || expertIsGated(unit->group)) {
	output = unit->totalInput;
	MoutputVar(unit) = MinputVar(unit);
    } else {
      Distribution	in, *out = &MtransformedDist(unit) ;
      in.mean = unit->totalInput ;
      in.variance = MinputVar(unit) ;
      transformDistribution(&in, out) ;
      output =  out->mean ;
      MoutputVar(unit) = out->variance ;
    }
    unit->output = output;
}
/********************************************************************/
static void 
ensembleTransferBack(unit) 
  Unit 	unit ;
{

  if (unit->group->type & OUTPUT || expertIsGated(unit->group)) {
    MinputVarDeriv(unit) = MoutputVarDeriv(unit);
    unit->inputDeriv = unit->outputDeriv ;
  } else {
    Distribution	*out = &MtransformedDist(unit) ;

    unit->inputDeriv = unit->outputDeriv*out->dMean_dMean
      + MoutputVarDeriv(unit)*out->dVariance_dMean ;
    MinputVarDeriv(unit) = unit->outputDeriv*out->dMean_dVariance
      + MoutputVarDeriv(unit)*out->dVariance_dVariance ;
  }
}
/**********************************************************************/

/*********************************************************************
 *	Name:		ensembleSumSquareError(Unit)
 *	Description:	calculates the sum square error of a unit
 *			(the derivative of the sum square error)
 *                      Adds on a term for the outputVarDeriv
 *	Parameters:
 *	  	Unit	unit -  the unit (unchanged)
 *	Return Value:
 *	  static void	ensembleSumSquareError -  the error (derivative)
 *********************************************************************/
#define K	(-0.91893853)	/* ln(sqrt(2*PI)) */
/********************************************************************/
static void 
ensembleSumSquareError(unit) 
  Unit		unit ;
{
  Real	stdDev   = variableGetValue(MerrorStdDevVar(unit->group)) ;
  Real	variance = square(stdDev) ;
  Real	term ;

  term = (MoutputVar(unit) + square(MstanDevOfTargets(unit->net))
	  + square(unit->output - adjustedTarget(unit)))/variance ;

  unit->net->error += K + log(stdDev) + term/2.0 ;
}
/********************************************************************/
static void 
ensembleSumSquareErrorDeriv(unit) 
  Unit 	unit ;
{
  Real	stdDev	 = variableGetValue(MerrorStdDevVar(unit->group)) ;
  Real	variance = square(stdDev) ;
  Real	term ;
    
  term = (MoutputVar(unit) + square(MstanDevOfTargets(unit->net))
	  + square(unit->output - adjustedTarget(unit)))/variance ;

  variableAddToDeriv(MerrorStdDevVar(unit->group), (1 - term)/stdDev) ;

  MoutputVarDeriv(unit) += 1.0/(2.0*variance) ;

  unit->outputDeriv += (unit->output - adjustedTarget(unit))/variance ;
}
/********************************************************************/




/***********************************************************************
 *	Name:		unitFields
 *	Description:	returns an array of all the names of the fields
 *			in a unit.
 *	Parameters:	NONE
 *	Return Value:	as above
 ***********************************************************************/
String *
unitFields () {
  static char	*fields[] =  { 
  "Output",
  "Output Variance",
  "totalInput",
  "Input Variance",
  "target",
  "inputDeriv",
  "Input Variance Deriv",
  "outputDeriv",
  "Output Variance Deriv",
  NULL } ;

  return fields ;
}
/**********************************************************************/
Real 
unitValue(unit, fieldIdx, time)
  Unit		unit ;
  int		fieldIdx ;
  int		time ;
{
  Real	value ;

  switch (fieldIdx) {
  case 0:
    if (unit->net->type & RECURRENT)
      value = unit->outputHistory[time] ;
    else
      value = unit->output ;
    break ;
  case 1:
    value = MoutputVar(unit) ;
    break ;
  case 2:
    value = unit->totalInput ;
    break ;
  case 3:
    value = MinputVar(unit) ;
    break ;
  case 4:
    if (unit->net->type & RECURRENT)
      value = unit->targetHistory[time] ;
    else
      value = unit->target ;
    break ;
  case 5:
    value = unit->inputDeriv ;
    break ;
  case 6:
    value = MinputVarDeriv(unit) ;
    break ;
  case 7:
    value = unit->outputDeriv ;
    break ;
  case 8:
    value = MoutputVarDeriv(unit) ;
    break ;
  default:
    value = 0.0 ;
    break ;
  }
  return value ;
}
/**********************************************************************/
void 
setUnitValue(unit, fieldIdx, time, value)
  Unit		unit ;
  int		fieldIdx ;
  int		time ;
  double	value ;
{
  switch (fieldIdx) {
  case 0:
    if (unit->net->type & RECURRENT)
      unit->outputHistory[time] = value ;
    else
      unit->output = value ;
    break ;
  case 1:
    MoutputVar(unit) = value ;
    break ;
  case 2:
    unit->totalInput = value ;
    break ;
  case 3:
    MinputVar(unit) = value ;
    break ;
  case 4:
    if (unit->net->type & RECURRENT)
      unit->targetHistory[time] = value ;
    else
      unit->target = value ;
    break ;
  case 5:
    unit->inputDeriv = value ;
    break ;
  case 6:
    MinputVarDeriv(unit) = value ;
  case 7:
    unit->outputDeriv = value ;
    break ;
  case 8:
    MoutputVarDeriv(unit) = value ;
    break ;
  default:
    break ;
  }
}
/**********************************************************************/

/***********************************************************************
 *	Name:		linkFields
 *	Description:	returns an array of all the names of the fields
 *			in a link.
 *	Parameters:	NONE
 *	Return Value:	as above
 ***********************************************************************/
String *
linkFields () {
  static char	*fields[] =  { 
  "weight",
  "Weight Variance",
  "Weight Cost",
  "deltaWeight",
  "deriv",
  NULL } ;

  return fields ;
}
/**********************************************************************/


/*********************************************************************
 *	Name:		linkValue
 *	Description:	returns the value of the requested field in a link
 *	Parameters:
 *	  Link		link - the link
 *	  int		fieldIdx - the index of the field (from "linkFields")
 *	Return Value:
 *	  static Real	linkValue - the value
 *********************************************************************/
Real 
linkValue(link, fieldIdx)
  Link	link ;
  int	fieldIdx ;
{
  Real	value ;

  switch (fieldIdx) {
  case 0:
    value = MgetWeight(link) ;
    break ;
  case 1:
    if (link->extension && MlinkVariance(link))
      value = MgetWeightVar(link) ;
    else
      value = 0.0 ;
    break ;
  case 2:
    if (link->extension)
      value = MgetLinkCost(link) ;
    else
      value = 0.0 ;
    break ;
  case 3:
    value = link->deltaWeight ;
    break ;
  case 4:
    value = link->deriv ;
    break ;
  default:
    value = 0 ;
    break ;
  }
  return value ;
}
/********************************************************************/
void 
setLinkValue(link, fieldIdx, value)
  Link		link ;
  int		fieldIdx ;
  double	value ;
{
  switch (fieldIdx) {
  case 0:
    variableSetValue(link, value) ;
    break ;
  case 1:
    if (link->extension && MlinkVariance(link))
      variableSetValue(MlinkVariance(link), value) ;
    break ;
  case 2:
    if (link->extension)
      MsetLinkCost(link, value) ;
    break ;
  case 3:
    link->deltaWeight = value ;
    break ;
  case 4:
    link->deriv = value ;
    break ;
  default:
    break ;
  }
}
/********************************************************************/


/***********************************************************************
 *	Name:		command_randomize
 *	Description:	command for randomizing network weights, also
 *			resets the currentEpoch
 *	Parameters:	
 *		int	tokc    - the number of command line tokens
 *		char	*tokv[] - the vector of tokens
 *	Return Value:	
 *		int	command_randomize - 0 on failure, 1 on success
 ***********************************************************************/
static Real
gaussianSample(this)
  Gaussian	this ;
{
  static Boolean	iset = FALSE ;
  static Real		gset ;
  Real			sample ;

  if  (iset) {
    sample = MGmean(this) + gset*MGstdDeviation(this) ;
    iset   = FALSE ;
  } else {
    Real	fac, r, v1, v2;
    do {
      v1 = -1.0 + 2.0*IRandReal() ;
      v2 = -1.0 + 2.0*IRandReal() ;
      r  = v1*v1 + v2*v2;
    } while (r >= 1.0);

    fac = sqrt(-2.0*log(r)/r);

    sample = MGmean(this) + v2*fac*MGstdDeviation(this) ;
    gset   = v1*fac ;
    iset   = TRUE ;
  }

  return sample ;
}
/********************************************************************/
static Real		stdDev, mean ;
/**********************************************************************/
static Real 
randReal(centre, field)
  double	centre ;
  double	field ;
{
  return centre + (2.0*IRandReal() - 1.0)*field ;
}
/**********************************************************************/
static void 
resetGroupCost(group, data)
  Group group ;
  void  *data ;
{
  resetEnsembleCost(NULL, group, mean, stdDev*sqrt(2)) ;
}
/********************************************************************/
int command_randomizeEnsemble (tokc, tokv)
  int	tokc ;
  char	*tokv[] ;
{
  int		idx, numLinks ;
  Gaussian	gaussian ;
  char		*name ;
  Boolean	uniform = FALSE ;

  IUsage("[-uniform] [<stdDev>] [<mean>]");
  if (GiveHelp(tokc)) {
    ISynopsis("Randomize weights on ensemble links in current net");
    IHelp
      (IHelpArgs,
       "Set all the weights to random  numbers with  a standard deviation of",
       "<stdDev> and a mean of <mean>.  Also  set the standard deviations of",
       "the weights to be <stdDev>. Default  values are mean = 0.0, stdDev =",
       "0.3.",
       "",
       "If the \"-uniform\"  option  is given,  the means  are  sampled from a",
       "uniform distribution ranging on:",
       "",
       "  [<mean> - <stdDev>, <mean> + <stdDev>]",
       "",
       "As well, if an ensemble cost model is being used, its prior mean and",
       "prior standard deviation  are set to  <mean>, and  sqrt(2)*<stdDev>,",
       "respectively.",
       "",
       "This command  also  clears the deltaWeight field  of  the  links and",
       "resets the net's currentEpoch.",
       "",
       "This command  does not change the weights  in links which have  been",
       "frozen, and it obeys all constraints placed on links.",
       "EXAMPLE",
       "\txerion-> randomize 1.0",
       "SEE ALSO",
       "randomize",
       NULL);
    return 1;
  }

  mean   = 0.0 ;
  stdDev = 0.3 ;

  name = *tokv ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-uniform", strlen(*tokv)) == 0) {
      uniform = TRUE ;
    } else if (*tokv[0] == '-') {
      IErrorAbort("Invalid option: %s", *tokv) ;
    } else {
      break ;
    }
  }

  if (tokc) {
    stdDev = atof(*tokv) ;
    ++tokv, --tokc ;
  }

  if (tokc) {
    mean   = atof(*tokv) ;
    ++tokv, --tokc ;
  }

  if (tokc) {
    IErrorAbort(IPrintUsage(name, usage));
    return 0 ;
  }

  if (currentNet == NULL)
    IErrorAbort("No current network") ;

  gaussian = createGaussian(NULL, mean, stdDev*stdDev) ;
  gaussian->getSample = gaussianSample ;

  numLinks = currentNet->numLinks - currentNet->numFrozenLinks ;
  for (idx = 0 ; idx < numLinks ; ++idx) {
    Link	link = currentNet->links[idx] ;
    if (link->type & ENSEMBLELINK) {
      if (uniform)
	link->weight = randReal(mean, stdDev) ; 
      else
	link->weight = MGsample(gaussian) ;

      link->deriv = 0.0 ;

      link = MlinkVariance(link) ;
      link->weight = stdDev*stdDev ;
      link->deriv  = 0.0 ;
    }
  }
  MGdestroy(gaussian) ;
  syncValues(currentNet, VectorFromWeights) ;

  resetEnsembleCost(currentNet, NULL, mean, stdDev*sqrt(2)) ;
  
  netForAllGroups(currentNet, ALL, resetGroupCost, NULL) ;

  currentNet->currentEpoch = 0 ;

  markToRebuildDisplay(CONNECTION_DISPLAY) ;
  return 1 ;
}
/**********************************************************************/
