#include <math.h>
#include <xerion/useful.h>
#include <xerion/commands.h>
#include <xerion/simulator.h>

#include "unit.h"
#include "ensemble.h"
#include "mixtureCost.h"

#define M_SQR(x)		((x)*(x))

/**********************************************************************
 * If you have to have any internal state, stick it in this structure.
 * Use MCMfoo as a template on how to access the variable.
 **********************************************************************/
typedef struct EnsembleMixData {
  MixtureDataRec parentData	;
  unsigned int	n ;
  Real		propSum ;
  Variable	*prop ;
  Variable	*mean ;
  Variable	*stdDev ;
} *EnsembleMixData ;

#define MCMensembleMixData(x,field)	\
  (((EnsembleMixData)((x)->costModelData.sumSquareData))->field)

#define MCMn(x)			MCMensembleMixData(x,n)
#define MCMpropSum(x)		MCMensembleMixData(x,propSum)
#define MCMpropArray(x)		MCMensembleMixData(x,prop)
#define MCMmeanArray(x)		MCMensembleMixData(x,mean)
#define MCMstdDevArray(x)	MCMensembleMixData(x,stdDev)
#define MCMprop(x,i)		MCMpropArray(x)[i]
#define MCMmean(x,i)		MCMmeanArray(x)[i]
#define MCMstdDev(x,i)		MCMstdDevArray(x)[i]
#define MCMnet(x)		((x)->net)
#define MCMnumLinks(x)		((x)->numLinks)
#define MCMlink(x, i)		((x)->link[i])

#define MCMparentData(x)	MCMensembleMixData(x,parentData)

static void	syncParentData ARGS((CostModel	this)) ;

/**********************************************************************
 * These are the procedures for getting things from the cost model:
 *
 *  MCMnet(c)		- returns the net containing the cost model
 *  MCMnumLinks(c)	- returns the number of variables in the model
 *  MCMlink(c, i) 	- returns the i'th variable ;
 *
 *  These are the procedures for manipulating variables:
 *
 *  Real variableGetValue(v) 	 - returns the value
 *  Real variableGetDeriv(v) 	 - returns the derivative
 *  char *variableGetName(v) 	 - returns its name
 *  Mask variableGetType(v)  	 - returns its type mask
 *  Net  variableGetNet(v)   	 - returns the net the variable belongs to
 *  Real variableSetValue(v,x)	 - sets the value of the variable to x
 *  Real variableSetDeriv(v,x)	 - sets the derivative of the variable to x
 *  real variableAddToDeriv(v,x) - adds x to the derivative of x
 **********************************************************************/

/**********************************************************************
 * This method should reset the cost model to its default values
 * (e.g.  it should reset means, variances, etc.)
 **********************************************************************/
static void 
reset(this, min, max)
  CostModel	this ;
  double	min ;
  double	max ;
{
  unsigned int	idx ;
  unsigned int	n = MCMn(this) ;
  Real		interval = (max - min)/n ;
  Real		mean     = min + interval/2.0 ;
  Real		stdDev	 = interval/2.0 ;

  /* reset the variables */
  for (idx = 0 ; idx < n ; ++idx) {
    Variable	variable ;
    
    variable = MCMprop(this, idx) ;
    if (!(variableGetType(variable) & FROZEN))
      variableSetValue(variable,	1.0/n) ;

    variable = MCMmean(this, idx) ;
    if (!(variableGetType(variable) & FROZEN))
      variableSetValue(variable,	mean) ;
    mean += interval ;

    variable = MCMstdDev(this, idx) ;
    if (!(variableGetType(variable) & FROZEN))
      variableSetValue(variable,	stdDev) ;
  }

  MCMsync(this) ;

  return ;
}
/**********************************************************************
 * This method should make sure that all internal variables have the
 * proper values for calculating the cost etc.
 **********************************************************************/
static void 
mysync(this)
  CostModel     this ;
{
  unsigned int	n   = MCMn(this) ;
  unsigned int	idx ;

  if (this->missingLinks)
    MCMcompact(this) ;

  /* normalize the proportions */
  MCMpropSum(this) = 0.0 ;
  for (idx = 0 ; idx < n ; ++idx)
    MCMpropSum(this) += variableGetValue(MCMprop(this, idx)) ;

  if (MCMpropSum(this) != 1.0) {
    for (idx = 0 ; idx < n ; ++idx) {
      variableSetValue(MCMprop(this, idx), 
		       variableGetValue(MCMprop(this, idx))/MCMpropSum(this)) ;
    }
  }

  syncParentData(this) ;
}
/**********************************************************************
 * This method should calculate the cost and return its value.
 **********************************************************************/
static Real 
ensembleMixCost(this)
  CostModel     this ;
{
  Real		logSum, subSum ;
  unsigned int	n   = MCMn(this) ;
  Real		*j  = (Real *)calloc(n, sizeof(Real)) ;
  unsigned int	priorIdx, linkIdx ;

  MCMsync(this) ;

  logSum = 0.0 ;
  for (linkIdx = 0 ; linkIdx < MCMnumLinks(this) ; ++linkIdx) {
    Link	link	 = MCMlink(this, linkIdx) ;
    Real	mean	 = MgetWeight(link) ;
    Real	variance = MgetWeightVar(link) ;
    Real	stdDev	 = sqrt(variance) ;

    subSum = 0.0 ;
    for (priorIdx = 0 ; priorIdx < n ; ++priorIdx) {
      Real	priorProp   = variableGetValue(MCMprop(this, priorIdx)) ;
      Real	priorMean   = variableGetValue(MCMmean(this, priorIdx)) ;
      Real	diff	    = priorMean - mean ;
      Real	priorStdDev = variableGetValue(MCMstdDev(this, priorIdx)) ;
      Real	priorVariance = M_SQR(priorStdDev) ;
      Real	f ;

      f = (variance - priorVariance + M_SQR(diff))/(2.0*priorVariance) ;

      subSum += j[priorIdx] = (priorProp*stdDev/priorStdDev) * exp(-f) ;
    }

    MsetLinkCost(link, -this->scale*log(subSum)) ;
    logSum += MgetLinkCost(link) ;
  }
  free((char *)j) ;
 
 return logSum ;
}
/**********************************************************************
 * This method should calculate the cost and add all the derivatives
 * into the list of variables.
 **********************************************************************/
#define M_DELTA(i,j)	((i) == (j) ? 1 : 0)
static Real 
ensembleMixCostAndDerivs(this)
  CostModel     this ;
{
  Real		logSum, subSum ;
  unsigned int	n   = MCMn(this) ;
  Real		*j  = (Real *)calloc(n, sizeof(Real)) ;
  unsigned int	priorIdx, linkIdx ;

  MCMsync(this) ;

  logSum = 0.0 ;
  for (linkIdx = 0 ; linkIdx < MCMnumLinks(this) ; ++linkIdx) {
    Link	link	 = MCMlink(this, linkIdx) ;
    Real	mean	 = MgetWeight(link) ;
    Real	variance = MgetWeightVar(link) ;
    Real	stdDev	 = sqrt(variance) ;

    subSum = 0.0 ;
    for (priorIdx = 0 ; priorIdx < n ; ++priorIdx) {
      Real	priorProp   = variableGetValue(MCMprop(this, priorIdx)) ;
      Real	priorMean   = variableGetValue(MCMmean(this, priorIdx)) ;
      Real	diff	    = priorMean - mean ;
      Real	priorStdDev = variableGetValue(MCMstdDev(this, priorIdx)) ;
      Real	priorVariance = M_SQR(priorStdDev) ;
      Real	f ;

      f = (variance - priorVariance + M_SQR(diff))/(2.0*priorVariance) ;

      subSum += j[priorIdx] = (priorProp*stdDev/priorStdDev) * exp(-f) ;
    }

    MsetLinkCost(link, -this->scale*log(subSum)) ;
    logSum += MgetLinkCost(link) ;

    /* normalize j */
    for (priorIdx = 0 ; priorIdx < n ; ++priorIdx)
      j[priorIdx] /= subSum ;

    for (priorIdx = 0 ; priorIdx < n ; ++priorIdx) {
      Real	priorProp   = variableGetValue(MCMprop(this, priorIdx)) ;
      Real	priorMean   = variableGetValue(MCMmean(this, priorIdx)) ;
      Real	priorStdDev = variableGetValue(MCMstdDev(this, priorIdx)) ;
      Real	priorVariance = M_SQR(priorStdDev) ;
      Real	diff	 = priorMean - mean ;
      Real	ratio	 = variance/priorVariance ;
      int	subIdx ;
	
      MaddToWeightDeriv(link,		
			-this->scale*j[priorIdx]*(diff/priorVariance)) ;

      variableAddToDeriv(MCMmean(this, priorIdx),	 
			 this->scale*j[priorIdx]*(diff/priorVariance)) ;
      
      MaddToWeightVarDeriv(link,
			   this->scale*j[priorIdx]*(ratio - 1.0)/(2*variance)) ;

      variableAddToDeriv(MCMstdDev(this, priorIdx),
			 -this->scale
			 *(j[priorIdx]/priorStdDev)
			 *(ratio - 1 + M_SQR(diff)/priorVariance)) ;

      for (subIdx = 0 ; subIdx < n ; ++subIdx) {
	variableAddToDeriv(MCMprop(this, subIdx),
			   -this->scale
			   *j[priorIdx]
			   *(M_DELTA(subIdx, priorIdx) - priorProp)
			   /(priorProp*MCMpropSum(this))) ;
      }
    }
  }

  free((char *)j) ;

  return logSum ;
}
/**********************************************************************/

/**********************************************************************
 * DO NOT CHANGE ANYTHING AFTER THIS POINT
 **********************************************************************/

static CostModel createEnsembleMixCostModel ARGS((char *name, Net net, int n,
						  double min, double max)) ;
static void	 resetEnsembleMixCostModel ARGS((CostModel, double, double)) ;

#define CM_ENSEMBLEMIX		CM_MIXTURE

/**********************************************************************/
int 
command_addEnsembleMixCost(tokc,tokv)
  int	tokc ;
  char	*tokv[] ;
{
  CostModel	costModel ;
  Group		group ;
  String	name ;
  Real		min, max ;
  int		numGaussians ;

  IUsage("[-min <x>] [-max <y>] [-group <group>] <n>");
  if (GiveHelp(tokc)) {
    ISynopsis("set the network cost model to use mixture ensemble cost model") ;
    IHelp
      (IHelpArgs,
       "This command sets the cost model used by the current network  to use",
       "a mixture ensemble cost  model.  That  is, the network  cost will be",
       "the  network  field \"weightCost\"  times the  mixture  ensemble  cost",
       "(whatever that is).",
       "",
       "The \"-group\" option allows you to create a cost model that acts only",
       "on the incoming links of a  single group. If a  group does not  have",
       "it's own cost model, it uses the network cost model.",
       "",
       "SEE ALSO",
       "addEnsembleCost, addMixtureCost, addSumSquareCost,",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("No currentNet") ;

  name = *tokv ;
  min = -0.3 ;
  max =  0.3 ;
  group = NULL ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-min", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      if (!IIsNumber(*tokv))
	IErrorAbort(IPrintUsage(name, usage)) ;
      min = atof(*tokv) ;
    } else if (strncmp(*tokv, "-max", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      if (!IIsNumber(*tokv))
	IErrorAbort(IPrintUsage(name, usage)) ;
      max = atof(*tokv) ;
    } else if (strncmp(*tokv, "-group", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      group = groupFromName(currentNet, *tokv) ;
      if (group == NULL)
	IErrorAbort("Unknown group: \"%s\"", *tokv) ;
    } else if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }

  if (tokc != 1 || !IIsInteger(tokv[0])) 
    IErrorAbort(IPrintUsage(name, usage)) ;

  numGaussians = atol(tokv[0]) ;

  return addEnsembleMixCost(currentNet, group, numGaussians, min, max) ;
}
/**********************************************************************/
int	
addEnsembleMixCost(net, group, num, min, max)
  Net	net ;
  Group	group ;
  int 	num ;
  double min ;
  double max ;
{
  CostModel	costModel ;

  if (group) {
    if (McostModel(group))
      MCMdestroy(McostModel(group)) ;

    McostModel(group) 
      = createEnsembleMixCostModel(group->name, net, num, min, max) ;

  } else {
    if (McostModel(net))
      MCMdestroy(McostModel(net)) ;

    McostModel(net) 
      = createEnsembleMixCostModel(net->name, net, num, min, max) ;
  }

  markToRebuildDisplay(ALL_DISPLAYS) ;

  return 1 ;
}
/**********************************************************************/


/*******************************************************************
 *	Name:		resetMixtureCost
 *	Description:	
 *	Parameters:
 *	  int	tokc - 
 *	  char	**tokv - 
 *	Return Value:
 *	  int	command_resetMixtureCost - 
 *******************************************************************/
int
command_resetEnsembleMixCost(tokc,tokv)
  int	tokc ;
  char	*tokv[] ;
{
  CostModel	costModel ;
  Group		group ;
  String	name ;
  Real		min, max ;

  IUsage("[-min <x>] [-max <y>] [-group <group>]");
  if (GiveHelp(tokc)) {
    ISynopsis("reset the mixture ensemble of gaussians cost model in the net");
    IHelp
      (IHelpArgs,
       "This command is used to reset the mixture ensemble cost model in the",
       "current net (see addEnsembleCost for a description of this model).",
       "",
       "The \"-group\" option  allows you to reset the cost model of  a single",
       "group.",
       NULL) ;
    return 1 ;
  }

  if (currentNet == NULL)
    IErrorAbort("No currentNet") ;

  name = *tokv ;
  min = -1.0 ;
  max =  1.0 ;
  group = NULL ;
  for (++tokv, --tokc ; tokc > 0 ; ++tokv, --tokc) {
    if (strncmp(*tokv, "-min", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      if (!IIsNumber(*tokv))
	IErrorAbort(IPrintUsage(name, usage)) ;
      min = atof(*tokv) ;
    } else if (strncmp(*tokv, "-max", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      if (!IIsNumber(*tokv))
	IErrorAbort(IPrintUsage(name, usage)) ;
      max = atof(*tokv) ;
    } else if (strncmp(*tokv, "-group", strlen(*tokv)) == 0) {
      ++tokv, --tokc ;
      group = groupFromName(currentNet, *tokv) ;
      if (group == NULL)
	IErrorAbort("Unknown group: \"%s\"", *tokv) ;
    } else if (*tokv[0] == '-') {
      IErrorAbort(IPrintUsage(name, usage)) ;
    } else {
      break ;
    }
  }

  if (tokc != 0)
    IErrorAbort(IPrintUsage(name, usage)) ;

  if (resetEnsembleMixCost(currentNet, group, min, max) == 0)
    IErrorAbort("Not using ensemble of gaussians cost model") ;

  return 1 ;
}
/**********************************************************************/
int 
resetEnsembleMixCost(net, group, min, max)
  Net	net ;
  Group	group ;
  double min ;
  double max ;
{
  CostModel	costModel ;

  if (group)
    costModel = McostModel(group) ;
  else
    costModel = McostModel(currentNet) ;

  if (costModel == NULL || MCMtype(costModel) != CM_ENSEMBLEMIX)
    return 0 ;

  resetEnsembleMixCostModel(costModel, min, max) ;

  markToRebuildDisplay(ALL_DISPLAYS) ;

  return 1 ;
}
/**********************************************************************/
static void 
resetEnsembleMixCostModel(this, min, max)
  CostModel this ;
  double min ;
  double max ;
{
  if (this->type == CM_ENSEMBLEMIX)
    this->reset(this, min, max) ;
}
/********************************************************************/


/**********************************************************************/
static void 
createParentData(this, min, max)
  CostModel	this ;
  double	min ;
  double	max ;
{
  MixtureData	parentData = &MCMparentData(this) ;
  unsigned int	n = MCMn(this) ;

  parentData->mean	   = NULL ;
  parentData->stdDeviation = NULL ;
  parentData->proportion   = NULL ;
  parentData->numGaussians = n ;
  parentData->mixture	   = createMixture(n, min, max) ;

  syncParentData(this) ;
}
/**********************************************************************/
static void 
destroyParentData(this)
  CostModel     this ;
{
  MixtureData	parentData = &MCMparentData(this) ;

  if (parentData->mixture)
    MMdestroy(parentData->mixture) ;
}
/**********************************************************************/
static void 
syncParentData(this)
  CostModel     this ;
{
  MixtureData	parentData = &MCMparentData(this) ;
  unsigned int	n = MCMn(this) ;
  unsigned int	idx ;

  for (idx = 0 ; idx < n ; ++idx) {
    Gaussian	gaussian = MMgaussian(parentData->mixture, idx) ;

    MGsetMean	    (gaussian,	variableGetValue(MCMmean  (this, idx))) ;
    MGsetVariance   (gaussian,  M_SQR(variableGetValue(MCMstdDev(this, idx))));
    MGsetProportion (gaussian,	variableGetValue(MCMprop  (this, idx))) ;
  }
}
/**********************************************************************/

/**********************************************************************
 * This method should destroy an ensemble mixtur model
 **********************************************************************/
static void 
destroy(this)
  CostModel     this ;
{
  unsigned int	idx ;
  unsigned int	n = MCMn(this) ;

  destroyParentData(this) ;

  if (MCMpropArray(this)) {
    for (idx = 0 ; idx < n ; ++idx)
      if (MCMprop(this, idx))
	destroyVariable(MCMprop(this, idx), MCMnet(this)) ;
  }

  if (MCMmeanArray(this)) {
    for (idx = 0 ; idx < n ; ++idx)
      if (MCMmean(this, idx))
	destroyVariable(MCMmean(this, idx), MCMnet(this)) ;
  }

  if (MCMstdDevArray(this)) {
    for (idx = 0 ; idx < n ; ++idx)
      if (MCMstdDev(this, idx))
	destroyVariable(MCMstdDev(this, idx), MCMnet(this)) ;
  }

  if (this->costModelData.sumSquareData)
    free((void *)this->costModelData.sumSquareData) ;
}
/**********************************************************************/
static CostModel 
createEnsembleMixCostModel(name, net, n, min, max)
  char	*name ;
  Net	net ;
  int 	n ;
  double min ;
  double max ;
{
  CostModel	this = createSumSquareCostModel(name, net) ;
  unsigned int	idx ;

  if (this == NULL)
    return NULL ;
  
  MCMtype(this) = CM_ENSEMBLEMIX ;
  
  this->destroy			= destroy ;
  this->reset			= reset ;
  this->sync			= mysync ;
  this->evaluateCost		= ensembleMixCost ;
  this->evaluateCostAndDerivs	= ensembleMixCostAndDerivs ;

  /* This is the messy part */
  if (this->costModelData.sumSquareData)
    free((void *)this->costModelData.sumSquareData) ;

  this->costModelData.sumSquareData = 
    (SumSquareData)malloc(sizeof(struct EnsembleMixData)) ;

  if (this->costModelData.sumSquareData == NULL)
    return NULL ;

  MCMn(this) = n ;

  MCMpropArray(this)   = (Variable *)calloc(n, sizeof(Variable)) ;
  MCMmeanArray(this)   = (Variable *)calloc(n, sizeof(Variable)) ;
  MCMstdDevArray(this) = (Variable *)calloc(n, sizeof(Variable)) ;

  for (idx = 0 ; idx < n ; ++idx) {
    char	buffer[128] ;
    sprintf(buffer, "%s.Proportion.%d", name, idx) ;
    MCMprop(this,idx)	= createVariable(buffer, net, LOG_TRANSFORM) ;
    deleteLinkFromCostModel(MCMprop(this, idx), NULL, net) ;

    sprintf(buffer, "%s.Mean.%d", name, idx) ;
    MCMmean(this,idx)	= createVariable(buffer, net, UNKNOWN) ;
    deleteLinkFromCostModel(MCMmean(this, idx), NULL, net) ;

    sprintf(buffer, "%s.StdDeviation.%d", name, idx) ;
    MCMstdDev(this,idx)	= createVariable(buffer, net, LOG_TRANSFORM) ;
    deleteLinkFromCostModel(MCMstdDev(this, idx), NULL, net) ;
  }

  createParentData(this, min, max) ;
  this->reset(this, min, max) ;

  return this ;
}
/**********************************************************************/
